/*
 * Copyright (C) 2002-2003 Stefan Stiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.kiwisoft.scrap;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;
import javax.swing.*;
import javax.swing.event.DocumentEvent;
import javax.swing.text.*;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.ProjectRootManager;
import com.intellij.openapi.vfs.VirtualFile;

import com.kiwisoft.utils.DocumentAdapter;
import com.kiwisoft.utils.StringUtils;
import com.kiwisoft.utils.Utils;
import com.kiwisoft.utils.format.FormatManager;
import com.kiwisoft.utils.gui.ConfirmationDialogStub;
import com.kiwisoft.utils.gui.IconManager;
import com.kiwisoft.utils.gui.inspect.InspectorDialogStub;
import com.kiwisoft.utils.gui.text.SourceTextPane;
import com.kiwisoft.utils.gui.text.SyntaxDefinitionFactory;
import net.sf.pizzacompiler.compiler.*;

/**
 * @author Stefan Stiller
 * @version $Revision: 1.1 $, $Date: 2004/10/16 18:21:39 $
 */
public class ScrapPanel extends JPanel
{
	public static final int ERROR=1;
	public static final int OUTPUT=2;
	public static final int RESULT=3;
	public static final int INFO=4;
	public static final int DEFAULT=0;

	private Project project;
	private MainPanel mainPanel;
	private Scrap scrap;
	private ScrapResult scrapResult;

	private JTextComponent tfScrap;
	private JTextPane tfResult;
	private StatusBar pnlStatus;

	private ScrapDocumentListener scrapDocumentListener;

	private RunAction runAction;
	private EvaluateAction evaluateAction;
	private InspectAction inspectAction;

	private ExecutionThread executionThread;

	public ScrapPanel(Project project, MainPanel mainPanel, Scrap scrap)
	{
		this.project=project;
		this.mainPanel=mainPanel;
		this.scrap=scrap;
		setLayout(new BorderLayout());

		runAction=new RunAction();
		evaluateAction=new EvaluateAction();
		inspectAction=new InspectAction();

		tfScrap=new SourceTextPane(SyntaxDefinitionFactory.getInstance().getSyntaxDefinition("java"));

		if (scrap!=null && scrap.getText()!=null) tfScrap.setText(scrap.getText());
		scrapDocumentListener=new ScrapDocumentListener();
		tfScrap.getDocument().addDocumentListener(scrapDocumentListener);

		JScrollPane pnlQuery=new JScrollPane(tfScrap);

		JPanel pnlButtons=new JPanel(new GridBagLayout());
		pnlButtons.add(Utils.createButton(runAction),
				new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Utils.createButton(evaluateAction),
				new GridBagConstraints(0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Utils.createButton(inspectAction),
				new GridBagConstraints(0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Utils.createButton(new CancelAction()),
				new GridBagConstraints(0, 3, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Box.createGlue(),
				new GridBagConstraints(0, 6, 1, 1, 0.0, 1.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));

		tfResult=new JTextPane();
		Style style=tfResult.addStyle("error", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.RED);
		style=tfResult.addStyle("info", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.GREEN);
		style=tfResult.addStyle("default", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.BLACK);
		style=tfResult.addStyle("output", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.BLACK);
		style=tfResult.addStyle("result", null);
		style.addAttribute(StyleConstants.ColorConstants.Bold, Boolean.TRUE);

		pnlStatus=new StatusBar();
		pnlStatus.setPreferredSize(new Dimension(20, 20));

		JSplitPane splitPane=new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		splitPane.setOneTouchExpandable(true);
		splitPane.setTopComponent(pnlQuery);
		splitPane.setBottomComponent(new JScrollPane(tfResult));
		splitPane.setDividerLocation(100);

		JPanel pnlMain=new JPanel(new BorderLayout());
		pnlMain.add(splitPane, BorderLayout.CENTER);
		pnlMain.add(pnlStatus, BorderLayout.SOUTH);

		add(pnlButtons, BorderLayout.WEST);
		add(pnlMain, BorderLayout.CENTER);
	}

	public void addNotify()
	{
		super.addNotify();
		validateActions();
	}

	public Scrap getScrap()
	{
		return scrap;
	}

	private void validateActions()
	{
		if (runAction!=null) runAction.validate();
		if (evaluateAction!=null) evaluateAction.validate();
		if (inspectAction!=null) inspectAction.validate();
	}

	private class ScrapDocumentListener extends DocumentAdapter
	{
		public void changedUpdate(DocumentEvent e)
		{
			scrap.setText(tfScrap.getText());
			validateActions();
		}
	}

	private String getScrapText()
	{
		pnlStatus.showMessage("");

		String fullText=tfScrap.getText();
		String partialText=tfScrap.getSelectedText();

		String scrapText=partialText;

		if (StringUtils.isEmpty(scrapText))
			scrapText=fullText;
		else
		{
			ScrapConfig config=ScrapConfig.getInstance(project);
			Boolean confirm=config.getConfirmation("partialExecute");
			if (Boolean.FALSE.equals(confirm))
			{
				scrapText=fullText;
			}
			else if (confirm==null)
			{
				ConfirmationDialogStub dialog=new ConfirmationDialogStub("Question", "Do you want to execute only the selected part of the scrap?",
						new String[]{"Yes", "No", "Cancel"}, "Yes", true);
				Utils.wrapDialog(project, dialog, true, true);
				if ("Yes".equals(dialog.getReturnValue()))
				{
						if (dialog.isDontShowAgain()) config.setConfirmation("partialExecute", Boolean.TRUE);
				}
				else if ("No".equals(dialog.getReturnValue()))
				{
						if (dialog.isDontShowAgain()) config.setConfirmation("partialExecute", Boolean.FALSE);
					scrapText=fullText;
				}
				else
					return null;
			}
		}

		return scrapText.trim();
	}

	protected void evaluateScrap()
	{
		if (executionThread==null || !executionThread.isAlive())
		{
			String scrapText=getScrapText();
			if (scrapText!=null)
			{
				String[] lines=scrapText.split("\r|\n");
				StringBuffer scrapCode=new StringBuffer();
				for (int i=0; i<lines.length; i++)
				{
					String line=lines[i];
					if (i==lines.length-1)
						scrapCode.append("return $result("+line+");");
					else
						scrapCode.append(line);
					scrapCode.append("\n");
				}
				executeScrap(scrapCode.toString());
			}
		}
	}

	protected void runScrap()
	{
		if (executionThread==null || !executionThread.isAlive())
		{
			String scrapText=getScrapText();
			if (scrapText!=null)
			{
				executeScrap(scrapText+"return null;");
			}
		}
	}

	private void executeScrap(String scrapCode)
	{
		synchronized (ExecutionThread.class)
		{
			try
			{
				net.sf.pizzacompiler.compiler.Main.init();
				ClassLoader classLoader;
				ClassReader classReader;
				if (project==null)
				{
					classLoader=Thread.currentThread().getContextClassLoader();
					classReader=new ClassReader();
				}
				else
				{
					Set projectClassPath=getProjectClassPath();

					StringBuffer classPath=new StringBuffer();
					URL[] urls=new URL[projectClassPath.size()];
					int i=0;
					for (Iterator it=projectClassPath.iterator(); it.hasNext();i++)
					{
						String pathElement=(String)it.next();
						classPath.append(pathElement).append(File.pathSeparator);
						urls[i]=new File(pathElement).toURL();
					}
					classPath.append(Utils.getClassBasePath(FormatManager.class).getAbsolutePath()).append(File.pathSeparator);
					classPath.append(Utils.getClassBasePath(ScrapPlugin.class).getAbsolutePath()).append(File.pathSeparator);
					classPath.append(Utils.getClassBasePath(ClassReader.class).getAbsolutePath()).append(File.pathSeparator);

					classReader=new ClassReader(null, classPath.toString());
					classLoader=new URLClassLoader(urls, ScrapBase.class.getClassLoader());
				}
				executionThread=new ExecutionThread(scrapCode, mainPanel.getImports(), classLoader, classReader);
				clearLastResults();
				runAction.setEnabled(false);
				evaluateAction.setEnabled(false);
				pnlStatus.startProgress();
				executionThread.start();
			}
			catch (MalformedURLException e)
			{
				handleThrowable(e);
			}
		}
	}

//	private void executeStatement(String queryText, boolean complex)
//	{
//		// The classloader is created here because he has to be created in AWTEventThread
//		ClassLoader classLoader;
//		try
//		{
//			if (CommonConfiguration.getInstance(project).isIncludeProjectClasses())
//			{
//				classLoader=Utils.getProjectClassLoader(project);
//			}
//			else
//			{
//				classLoader=getClass().getClassLoader();
//			}
//		}
//		catch (Throwable e)
//		{
//			classLoader=getClass().getClassLoader();
//		}
//
//		StatementList statementList;
//		if (complex)
//			statementList=new StatementList(new ComplexSQLStatement(queryText));
//		else
//			statementList=SQLParser.parse(queryText);
//
//		executionThread=new ExecutionThread(statementList, classLoader);
//		initExecution(false);
//		executionThread.start();
//	}

//	public void executeStatement(final SQLStatement statement, final ResultSetPanel panel)
//	{
//		SwingUtilities.invokeLater(new Runnable()
//		{
//			public void run()
//			{
//				if (executionThread==null || !executionThread.isAlive())
//				{
//					pnlStatus.showMessage("");
//					errors=false;
//
//					ClassLoader classLoader;
//					try
//					{
//						classLoader=Utils.getProjectClassLoader(project);
//					}
//					catch (Throwable e)
//					{
//						classLoader=getClass().getClassLoader();
//					}
//					executionThread=new ExecutionThread(new StatementList(statement), classLoader);
//					executionThread.setResultSetPanel(panel);
//					initExecution(true);
//					executionThread.start();
//				}
//			}
//		});
//	}

	private class RunAction extends AbstractAction
	{
		public RunAction()
		{
			super("Run");
			putValue(Action.SMALL_ICON, IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/run.png"));
			putValue(Action.SHORT_DESCRIPTION, "Run scrap.");
			putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, KeyEvent.ALT_MASK));
		}

		public void validate()
		{
			setEnabled((executionThread==null || !executionThread.isAlive()) && !StringUtils.isEmpty(tfScrap.getText()));
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				runScrap();
			}
			catch (Exception e1)
			{
				handleThrowable(e1);
			}
		}
	}

	private class EvaluateAction extends AbstractAction
	{
		public EvaluateAction()
		{
			super("Evaluate");
			putValue(Action.SMALL_ICON, IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/evaluate.png"));
			putValue(Action.SHORT_DESCRIPTION, "Evaluate scrap.");
			putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, KeyEvent.CTRL_MASK));
		}

		public void validate()
		{
			setEnabled((executionThread==null || !executionThread.isAlive()) && !StringUtils.isEmpty(tfScrap.getText()));
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				evaluateScrap();
			}
			catch (Exception e1)
			{
				handleThrowable(e1);
			}
		}
	}

	private class InspectAction extends AbstractAction
	{
		public InspectAction()
		{
			super("Inspect");
			putValue(Action.SMALL_ICON, IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/browser.png"));
			putValue(Action.SHORT_DESCRIPTION, "Inspect scrap result.");
		}

		public void validate()
		{
			setEnabled(scrapResult!=null && scrapResult.isObject());
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				if (scrapResult!=null && scrapResult.isObject())
					Utils.wrapDialog(project, new InspectorDialogStub("Inspector", scrapResult.getValue()), false, true);
			}
			catch (Exception e1)
			{
				handleThrowable(e1);
			}
		}
	}

	public void handleThrowable(Throwable e)
	{
		e.printStackTrace();
		showMessage(e.getMessage(), ERROR, true);
	}

	private void clearLastResults()
	{
		scrapResult=null;
		tfResult.setText("");
	}

	public void showMessage(String message, int type, boolean nextLine)
	{
		StyledDocument document=tfResult.getStyledDocument();
		try
		{
			if (document.getLength()>0) document.insertString(document.getLength(), "\n", null);
			switch (type)
			{
				case OUTPUT:
					document.setLogicalStyle(document.getLength(), document.getStyle("output"));
					break;
				case RESULT:
					document.setLogicalStyle(document.getLength(), document.getStyle("result"));
					break;
				case ERROR:
					document.setLogicalStyle(document.getLength(), document.getStyle("error"));
					break;
				case INFO:
					document.setLogicalStyle(document.getLength(), document.getStyle("info"));
					break;
				default:
					document.setLogicalStyle(document.getLength(), document.getStyle("default"));
			}
			document.insertString(document.getLength(), message, null);
			if (nextLine) document.insertString(document.getLength(), "\n", null);
			tfResult.setCaretPosition(document.getLength());
		}
		catch (BadLocationException e)
		{
			e.printStackTrace();
		}
	}

	private class CancelAction extends AbstractAction
	{
		public CancelAction()
		{
			super("Close scrap", IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/cancel.png"));
			putValue(Action.SHORT_DESCRIPTION, "Close current scrap.");
		}

		public void actionPerformed(ActionEvent e)
		{
			ScrapManager.getInstance(project).removeScrap(scrap);
		}
	}

	public void insertString(final String text)
	{
		SwingUtilities.invokeLater(new Runnable()
		{
			public void run()
			{
				Document document=tfScrap.getDocument();
				try
				{
					int selectionStart=tfScrap.getSelectionStart();
					int selectionEnd=tfScrap.getSelectionEnd();
					int position=tfScrap.getCaretPosition();
					if (selectionStart<selectionEnd)
					{
						position=selectionStart;
						document.remove(selectionStart, selectionEnd-selectionStart);
					}
					document.insertString(position, text, null);
					tfScrap.requestFocus();
				}
				catch (BadLocationException e)
				{
					e.printStackTrace();
				}
			}
		});
	}

	private Set getProjectClassPath()
	{
		ProjectRootManager rootManager = (ProjectRootManager) project.getComponent(ProjectRootManager.class);
		VirtualFile[] virtualFiles = rootManager.getFullClassPath();
		Set list=new LinkedHashSet();
		int i;
		for (i = 0; i < virtualFiles.length; i++)
		{
			String path = virtualFiles[i].getPath();
			list.add(path.endsWith("!/") ? path.substring(0, path.length() - 2) : path);
		}
		return list;
	}

	private class ExecutionThread extends Thread
	{
		private String text;
		private String[] imports;
		private ClassLoader classLoader;
		private ClassReader classReader;

		public ExecutionThread(String text, String[] imports, ClassLoader classLoader, ClassReader classReader)
		{
			this.text=text;
			this.imports=imports;
			this.classLoader=classLoader;
			this.classReader=classReader;
		}

		public void run()
		{
			long time1=System.currentTimeMillis();
			try
			{
				ByteArrayOutputStream messageStream=new ByteArrayOutputStream();
				try
				{
					StringBuffer sourceCode=new StringBuffer();
					if (imports!=null)
					{
						for (int i=0; i<imports.length; i++) sourceCode.append("import "+imports[i]+";\n");
					}
					sourceCode.append("public class $Test extends "+ScrapBase.class.getName()+" {\n");
					sourceCode.append("public "+ScrapResult.class.getName()+" $scrap("+PrintWriter.class.getName()+" out) {\n");
					sourceCode.append(text);
					sourceCode.append("}\n");
					sourceCode.append("}\n");
					System.out.println("sourceCode = "+sourceCode);

					net.sf.pizzacompiler.compiler.Main.setClassReader(classReader);

					ByteArrayCompilerOutput compilerOutput=new ByteArrayCompilerOutput();
					HashMap classMap=new HashMap();
					classMap.put("$Test", sourceCode.toString());
					SourceReader reader=new MapSourceReader(classMap);
					PrintStream sysOut=new PrintStream(messageStream);
					net.sf.pizzacompiler.compiler.Main.compile(new String[]{"$Test"}, reader, compilerOutput, sysOut);
					ClassLoader scrapLoader=new ByteArrayClassLoader(classLoader, compilerOutput.getBytecode());
					Class aClass=scrapLoader.loadClass("$Test");
					ScrapBase scrap=(ScrapBase)aClass.newInstance();
					Writer outputWriter=new Writer()
					{
						public void write(char cbuf[], int off, int len)
						{
							showMessage(new String(cbuf, off, len), OUTPUT, false);
						}

						public void flush()
						{
						}

						public void close()
						{
						}
					};
					scrapResult=scrap.$scrap(new PrintWriter(outputWriter));
					if (scrapResult!=null) showMessage(scrapResult.toString(), RESULT, true);
					System.out.println();
				}
				catch (Throwable throwable)
				{
					String throwableString=throwable.toString();
					if (throwableString==null || !"java.lang.ClassNotFoundException: scrap".equals(throwableString))
						throwable.printStackTrace();
					throw new RuntimeException(new String(messageStream.toByteArray()));
				}
			}
			catch (Throwable e)
			{
				handleThrowable(e);
			}
			finally
			{
				final long time=System.currentTimeMillis()-time1;
				SwingUtilities.invokeLater(new Runnable()
				{
					public void run()
					{
						pnlStatus.appendMessage("Execution time: "+time+"ms");
						validateActions();
						pnlStatus.stopProgress();
					}
				});
			}
		}
	}
}
